import logging

import click

from hoststats.app import app


@click.group()
def top():
    pass


@top.command()
def start():
    logging.info("Starting hoststats")

    app.run(debug=True, host="0.0.0.0")


def main():
    top()



import logging
from os.path import exists

import pandas as pd

from hoststats.stats import CPU_STATS, DISK_STATS, MEM_STATS, NET_STATS


def validate_csv_data(csv_path, expected_hosts):
    if not exists(csv_path):
        logging.error(f"Did not find hoststats file at {csv_path}")
        return False

    data = pd.read_csv(csv_path)

    is_valid = True

    # Check columns are as expected
    expected_cols = (
        ["Timestamp", "Host"] + CPU_STATS + MEM_STATS + DISK_STATS + NET_STATS
    )
    cols = [c for c in data.columns]
    if cols != expected_cols:
        logging.error(
            f"hoststats expected cols {expected_cols}, got {data.columns}"
        )
        is_valid = False

    # Check hosts list
    hosts = set(data["Host"].unique())
    if set(expected_hosts) != hosts:
        logging.error(
            f"hoststats invalid, expected hosts {expected_hosts}, got {hosts}"
        )
        is_valid = False

    if data.isnull().values.any():
        logging.error("Found null values in hoststats")
        is_valid = False

    return is_valid



import json
import logging
from subprocess import PIPE, run

import pandas as pd
import requests

from hoststats.stats import (
    CPU_STATS,
    DISK_STATS,
    FORWARD_HEADER,
    MEM_STATS,
    NET_STATS,
    SERVER_PORT,
)


class HostStats:
    def __init__(
        self,
        host_list,
        test_mode=False,
        proxy=None,
        proxy_port=SERVER_PORT,
        kubectl=False,
        kubectl_container=None,
        kubectl_ns=None,
    ):
        self.host_list = host_list
        self.test_mode = test_mode
        self.is_running = False
        self.proxy = proxy
        self.proxy_port = proxy_port

        self.kubectl = kubectl
        self.kubectl_container = kubectl_container
        self.kubectl_ns = kubectl_ns

        if self.test_mode:
            from hoststats.app import app

            self.client = app.test_client()

        successful_init = True

        for h in host_list:
            logging.debug(f"Pinging {h}")
            status, data = self.make_request(h, "ping")

            if status != 200:
                logging.error(f"Failed to ping {h}, got code {status}")
                successful_init = False

            if data != "PONG":
                logging.error(f"Got unexpected response to ping: {data}")
                successful_init = False

        if not successful_init:
            raise RuntimeError("hoststats client failed to initialise")

        if self.proxy:
            logging.debug(
                f"Created hoststats proxy on {self.proxy} for {self.host_list}"
            )
        else:
            logging.debug(f"Created hoststats client for {self.host_list}")

    def make_request(self, host, url):
        if self.test_mode:
            resp = self.client.get(url)
            status_code = resp.status_code

            data = None
            if resp.data:
                data = resp.data.decode("utf-8")

            return status_code, data

        if self.kubectl:
            cmd = [
                "kubectl",
                f"-n {self.kubectl_ns}" if self.kubectl_ns else "",
                f"-c {self.kubectl_container}"
                if self.kubectl_container
                else "",
                "exec",
                host,
                "--",
                f"curl -s http://localhost:5000/{url}",
            ]
            cmd_str = " ".join(cmd)
            res = run(cmd_str, shell=True, stdout=PIPE, stderr=PIPE)

            output = res.stdout.decode("utf-8")
            if res.returncode == 0:
                return 200, output
            else:
                return 500, output

        elif self.proxy:
            full_url = f"http://{self.proxy}:{self.proxy_port}/{url}"
            resp = requests.get(full_url, headers={FORWARD_HEADER: host})
        else:
            resp = requests.get(f"http://{host}:{SERVER_PORT}/{url}")

        return resp.status_code, resp.text

    def start_collection(self):
        successful_start = True

        for h in self.host_list:
            logging.debug(f"Starting collection on {h}")
            status, resp = self.make_request(h, "start")

            if status != 200:
                logging.error(
                    f"Failed to start on {h}, got code {resp.status_code}"
                )
                successful_start = False

            if not successful_start:
                raise RuntimeError("hoststats failed to start collection")

        self.is_running = True

    def stop_and_write_to_csv(self, csv_path):
        if not self.is_running:
            raise RuntimeError("Called stop without starting")

        self.is_running = False

        csv_cols = ["Timestamp", "Host"]
        csv_cols.extend(CPU_STATS)
        csv_cols.extend(MEM_STATS)
        csv_cols.extend(DISK_STATS)
        csv_cols.extend(NET_STATS)

        with open(csv_path, "w") as fh:
            fh.write(",".join(csv_cols))
            fh.write("\n")

        for h in self.host_list:
            logging.debug(f"Writing results for {h}")

            # Pull the results
            df = self.pull_results_for_host(h)

            # Append to the CSV file
            df.to_csv(csv_path, header=False, mode="a", index=False)

    def pull_results_for_host(self, host):
        status, data = self.make_request(host, "stop")
        data_json = json.loads(data)

        cpu_stats = pd.read_json(json.dumps(data_json["cpu"]), orient="list")
        mem_stats = pd.read_json(json.dumps(data_json["mem"]), orient="list")
        disk_stats = pd.read_json(json.dumps(data_json["disk"]), orient="list")
        net_stats = pd.read_json(json.dumps(data_json["net"]), orient="list")

        n_readings = len(cpu_stats["timestamp"])
        host_col = [host] * n_readings

        merged = pd.DataFrame(columns=["timestamp", "host"])
        merged["timestamp"] = cpu_stats["timestamp"]
        merged["host"] = host_col

        merged = pd.concat(
            [
                merged,
                cpu_stats[CPU_STATS],
                mem_stats[MEM_STATS],
                disk_stats[DISK_STATS],
                net_stats[NET_STATS],
            ],
            axis=1,
        )

        return merged



import logging
from os.path import exists

import pandas as pd


class HostStatsResults:
    def __init__(self, csv_file, relative_ts=True, base_ts=None):
        self.csv_file = csv_file

        if not exists(self.csv_file):
            logging.error(f"hoststats result file not found: {self.csv_file}")
            raise RuntimeError("hoststas result file not found")

        self.full_results = pd.read_csv(self.csv_file)

        # Parse timestamps to pd timestamp objects
        self.full_results["Timestamp"] = pd.to_datetime(
            self.full_results["Timestamp"]
        )

        # If no base timestamp given, take the first in the data
        if relative_ts:
            if not base_ts:
                base_ts = self.full_results["Timestamp"][0]

            # Put all timestamps relative to the base timestamp in seconds
            self.full_results["Timestamp"] -= base_ts

        # Merge across hosts for aggregate stats
        self.grouped_results = self.full_results.groupby("Timestamp")

    def get_hosts(self):
        return self.full_results["Host"].unique()

    def get_full_results(self):
        return self.full_results

    def get_grouped_results(self):
        return self.grouped_results

    def get_stats(self):
        stats = [c for c in self.full_results.columns]
        stats.remove("Timestamp")
        stats.remove("Host")

        return stats

    def validate_stat(self, stat):
        stats = self.get_stats()
        if stat not in stats:
            raise RuntimeError(f"{stat} not in {stats}")

    def get_stat_per_host(self, stat):
        self.validate_stat(stat)

        hosts = self.get_hosts()
        stats_per_host = dict()
        for host in hosts:
            host_ts = self.full_results["Host"] == host
            stats_per_host[host] = self.full_results.loc[host_ts][
                ["Timestamp", stat]
            ]

        return stats_per_host

    def get_avg_stat(self, stat):
        self.validate_stat(stat)
        return self.grouped_results.mean()[stat]

    def get_median_stat(self, stat):
        self.validate_stat(stat)
        return self.grouped_results.median()[stat]

    def get_stat_quantile(self, stat, q):
        self.validate_stat(stat)
        return self.grouped_results.quantile(q)[stat]



import logging
from multiprocessing import Process, Queue

import requests
from flask import Blueprint, Response, request

from hoststats.collection import collect_metrics
from hoststats.stats import FORWARD_HEADER, SERVER_PORT

metrics_api = Blueprint("metrics_api", __name__)

metrics_process = None
kill_queue = None
result_queue = None


def _get_forward_host():
    for key, value in request.headers.items():
        if key.lower() == FORWARD_HEADER.lower():
            return value

    return None


def _is_forward_request():
    host = _get_forward_host()
    return bool(host)


def _do_forward_request():
    # Note, Flask's request.host field contains the port too
    original_url = request.url
    target_host = _get_forward_host()
    forward_url = original_url.replace(
        request.host, f"{target_host}:{SERVER_PORT}"
    )

    # Strip out forward header and host header
    forward_headers = {
        k: v
        for (k, v) in request.headers
        if k.lower() not in [FORWARD_HEADER.lower(), "host"]
    }

    logging.debug(f"Forwarding request from {original_url} to {forward_url}")
    logging.debug(f"Forwarding headers: {forward_headers}")

    # Make the forward request
    resp = requests.request(
        method=request.method,
        url=forward_url,
        headers=forward_headers,
    )

    # Strip out undesired headers from the forwarded response
    excluded_headers = [
        "content-encoding",
        "content-length",
        "transfer-encoding",
        "connection",
    ]
    headers = [
        (name, value)
        for (name, value) in resp.headers.items()
        if name.lower() not in excluded_headers
    ]

    # Build Flask response from forwarded response
    return Response(resp.content, resp.status_code, headers)


@metrics_api.route("/ping")
def ping():
    if _is_forward_request():
        return _do_forward_request()

    return "PONG"


@metrics_api.route("/start")
def start_recording():
    if _is_forward_request():
        return _do_forward_request()

    global metrics_process
    global kill_queue
    global result_queue

    if metrics_process is not None:
        msg = "Not starting metrics recording, already running"
        logging.warn(msg)
        return msg

    kill_queue = Queue()
    result_queue = Queue()

    metrics_process = Process(
        target=collect_metrics, args=(kill_queue, result_queue)
    )
    metrics_process.start()

    return "hoststats started"


@metrics_api.route("/stop")
def stop_recording():
    if _is_forward_request():
        return _do_forward_request()

    global metrics_process
    global kill_queue
    global result_queue

    if metrics_process is None:
        msg = "Not stopping metrics recording, not running"
        logging.warn(msg)
        return msg

    kill_queue.put("die")
    # Get the results before we join the process, otherwise if the queue buffers
    # fill the program will deadlock
    result_json = result_queue.get()
    metrics_process.join()

    metrics_process = None
    kill_queue = None
    result_queue = None

    return result_json



import setuptools

PKG_NAME = "hoststats"

with open("VERSION", "r") as fh:
    pkg_ver = fh.read().strip()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name=PKG_NAME,
    version=pkg_ver,
    author="Simon Shillaker",
    author_email="foo@bar.com",
    description="Host stats",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Shillaker/hoststats",
    packages=["hoststats"],
    install_requires=[
        "click",
        "flask",
        "numpy",
        "pandas",
        "psutil",
        "requests",
    ],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["hoststats = hoststats.cli:main"]},
)



import pandas as pd

# Note - header keys must start with a capital, and be all lower-case otherwise
FORWARD_HEADER = "Forwardhost"
SERVER_PORT = "5000"

ONE_MB = 1024.0 * 1024.0

CPU_STATS = [
    "CPU_COUNT",
    "CPU_FREQ",
    "CPU_PCT",
    "CPU_TIME_IDLE",
    "CPU_TIME_IOWAIT",
    "CPU_TIME_USER",
    "CPU_TIME_SYSTEM",
    "CPU_PCT_IDLE",
    "CPU_PCT_IOWAIT",
    "CPU_PCT_USER",
    "CPU_PCT_SYSTEM",
]

MEM_STATS = [
    "MEMORY_USED",
    "MEMORY_ACTIVE",
    "MEMORY_USED_PCT",
    "MEMORY_AVAILABLE",
    "SWAP_USED",
]

DISK_STATS = ["DISK_READ_MB", "DISK_WRITE_MB"]

NET_STATS = ["NET_SENT_MB", "NET_RECV_MB"]


def get_stats_dfs():
    cpu_stats = pd.DataFrame(columns=["timestamp"] + CPU_STATS)

    mem_stats = pd.DataFrame(columns=["timestamp"] + MEM_STATS)

    disk_stats = pd.DataFrame(columns=["timestamp"] + DISK_STATS)

    net_stats = pd.DataFrame(columns=["timestamp"] + NET_STATS)

    return cpu_stats, mem_stats, disk_stats, net_stats



import json
import logging
import sys
import time
from os import environ
from time import sleep

import psutil

from hoststats.stats import ONE_MB, get_stats_dfs

SLEEP_INTERVAL_MS = int(environ.get("HOSTSTATS_SLEEP_INTERVAL_MS", "2000"))


def collect_metrics(kill_queue, result_queue):
    cpu_stats, mem_stats, disk_stats, net_stats = get_stats_dfs()

    # cpu_percent will return zeroes to start with, so we want to make an
    # initial call
    # https://psutil.readthedocs.io/en/latest/#psutil.cpu_percent
    psutil.cpu_percent(percpu=False, interval=None)
    psutil.cpu_percent(percpu=True, interval=None)

    # Get CPU count and add columns to df
    cpu_count = psutil.cpu_count()
    for cpu_num in range(cpu_count):
        cpu_stats[f"CPU_{cpu_num}_PCT"] = list()
        cpu_stats[f"CPU_{cpu_num}_TIME_IDLE"] = list()
        cpu_stats[f"CPU_{cpu_num}_TIME_IOWAIT"] = list()
        cpu_stats[f"CPU_{cpu_num}_TIME_USER"] = list()
        cpu_stats[f"CPU_{cpu_num}_TIME_SYSTEM"] = list()

        cpu_stats[f"CPU_{cpu_num}_PCT_IDLE"] = list()
        cpu_stats[f"CPU_{cpu_num}_PCT_IOWAIT"] = list()
        cpu_stats[f"CPU_{cpu_num}_PCT_USER"] = list()
        cpu_stats[f"CPU_{cpu_num}_PCT_SYSTEM"] = list()

    while True:
        # Check queue to see if we need to finish
        if not kill_queue.empty():
            res = kill_queue.get_nowait()

            if res:
                logging.debug("Metrics collection finished")

                full_data = {
                    "cpu": cpu_stats.to_dict(orient="list"),
                    "mem": mem_stats.to_dict(orient="list"),
                    "disk": disk_stats.to_dict(orient="list"),
                    "net": net_stats.to_dict(orient="list"),
                }

                result_queue.put(json.dumps(full_data))

                sys.stdout.flush()
                break

        timestamp = int(time.time() * 1000)

        # CPU
        cpu_pct = psutil.cpu_percent(percpu=False)
        cpu_times = psutil.cpu_times(percpu=False)
        cpu_times_pct = psutil.cpu_times_percent(percpu=False)

        cpu_pct_per_cpu = psutil.cpu_percent(percpu=True)
        cpu_times_per_cpu = psutil.cpu_times(percpu=True)
        cpu_times_pct_per_cpu = psutil.cpu_times_percent(percpu=True)

        cpu_row = [
            timestamp,
            psutil.cpu_count(),
            psutil.cpu_freq().max,
            cpu_pct,
            cpu_times.idle,
            cpu_times.iowait,
            cpu_times.user,
            cpu_times.system,
            cpu_times_pct.idle,
            cpu_times_pct.iowait,
            cpu_times_pct.user,
            cpu_times_pct.system,
        ]

        for cpu_num in range(cpu_count):
            cpu_row.extend(
                [
                    cpu_pct_per_cpu[cpu_num],
                    cpu_times_per_cpu[cpu_num].idle,
                    cpu_times_per_cpu[cpu_num].iowait,
                    cpu_times_per_cpu[cpu_num].user,
                    cpu_times_per_cpu[cpu_num].system,
                    cpu_times_pct_per_cpu[cpu_num].idle,
                    cpu_times_pct_per_cpu[cpu_num].iowait,
                    cpu_times_pct_per_cpu[cpu_num].user,
                    cpu_times_pct_per_cpu[cpu_num].system,
                ]
            )

        cpu_stats.loc[len(cpu_stats.index)] = cpu_row

        # MEMORY
        vmem = psutil.virtual_memory()
        swap = psutil.swap_memory()
        mem_row = [
            timestamp,
            vmem.used / ONE_MB,
            vmem.active / ONE_MB,
            vmem.percent,
            vmem.available / ONE_MB,
            swap.used / ONE_MB,
        ]

        mem_stats.loc[len(mem_stats.index)] = mem_row

        # DISK
        disk = psutil.disk_io_counters(perdisk=False)
        disk_row = [
            timestamp,
            disk.read_bytes / ONE_MB,
            disk.write_bytes / ONE_MB,
        ]
        disk_stats.loc[len(disk_stats.index)] = disk_row

        # NETWORK
        network = psutil.net_io_counters(pernic=False)
        net_row = [
            timestamp,
            network.bytes_sent / ONE_MB,
            network.bytes_recv / ONE_MB,
        ]
        net_stats.loc[len(net_stats.index)] = net_row

        # Sleep until next measurement
        sleep(float(SLEEP_INTERVAL_MS) / 1000)
